; Author: Petroizki
; an simple starfield example.

.686
.model flat, stdcall
	include windows.inc
	
	include gdi32.inc
	include kernel32.inc
	include masm32.inc
	include user32.inc
	
	includelib gdi32.lib
	includelib kernel32.lib
	includelib masm32.lib
	includelib user32.lib
	
	AMOUNT_OF_DOTS EQU <512>				; how many dots there are, this cannot be 0
	DOT_SPEED EQU <0.15>
	MINIMUM_LIFETIME EQU <30>				; minimum lifetime the dots have (in frames)
	REFRESH_DELAY EQU <33>					; ms, about 30fps
	SCREEN_HEIGHT EQU <480>					; window height
	SCREEN_WIDTH EQU <640>					; window width
	
	CreateDot proto :DWORD
	CreateDots proto
	main proto
	UpdateDots proto :HDC
	WndProc proto :DWORD, :DWORD, :DWORD, :DWORD
	
	POSITION2D struct
		x REAL4 ?
		y REAL4 ?
	POSITION2D ends
	
	SINGLE_DOT struct
		lifetime DWORD ?
		position POSITION2D <?>		; position of the dot
	SINGLE_DOT ends
.data?
	dots SINGLE_DOT AMOUNT_OF_DOTS dup (<?>)
.data
	g_dotspeed REAL4 DOT_SPEED
	g_winclass WNDCLASSEX <SIZEOF WNDCLASSEX, CS_HREDRAW or CS_VREDRAW, WndProc, 0, 0, NULL, NULL, NULL, \
		NULL, NULL, OFFSET STRING_WINCLASS, NULL>
	g_world RECT <?>
	g_worldcenter POSITION2D <?>			; center point of world, the point where viewer is going
	
	STRING_WINCLASS db "STARFIELD",0		; class name for createwindowex
	STRING_WINNAME db "Starfield",0			; window title
.code
start:
	invoke main
	invoke ExitProcess, 0

;------------------------------------------------------------------------------
; Program's main proc. Creates window, and handles the main loop.
;------------------------------------------------------------------------------
main proc uses ebx edi esi
	LOCAL msg:MSG
	LOCAL rect:RECT
	LOCAL temp:DWORD

	invoke GetStockObject, BLACK_BRUSH		; get background brush
	mov dword ptr [g_winclass.hbrBackground], eax
	invoke LoadCursor, NULL, IDC_ARROW		; get desktop cursor
	mov dword ptr [g_winclass.hCursor], eax
	invoke GetModuleHandle, NULL			; get process module handle
	mov dword ptr [g_winclass.hInstance], eax
	
	invoke RegisterClassEx, ADDR g_winclass	; register windowclass
	test eax, eax
	jz @force_exit							; fuck it
	
	invoke CreateWindowEx, 0, ADDR STRING_WINCLASS, ADDR STRING_WINNAME, WS_MINIMIZEBOX or WS_SYSMENU or WS_VISIBLE, \
		15, 20, SCREEN_WIDTH, SCREEN_HEIGHT, NULL, NULL, g_winclass.hInstance, NULL
	test eax, eax
	jz @force_exit							; terminate prog, it would be nice to have some error messages, but who cares
	
	mov ebx, eax							; store the window handle
	
	invoke GetDC, eax						; get window DC
	mov edi, eax
	
	invoke GetClientRect, ebx, ADDR g_world	; get window rectangle
	
	mov eax, g_world.right					; get right edge
	shr eax, 1								; divide it by 2 to get center point
	mov temp, eax
	fild dword ptr [temp]					; load center x-point to fpu register
	fstp dword ptr [g_worldcenter.x]		; store it as fpu
	
	mov eax, g_world.bottom					; get bottom edge
	shr eax, 1								; divide it by 2 to get center point
	mov temp, eax
	fild dword ptr [temp]					; load center y-point to fpu register
	fstp dword ptr [g_worldcenter.y]		; store it as fpu
	
	invoke nseed, 666						; prepare the nrandom
	invoke CreateDots
	
	invoke GetTickCount						; get current time
	lea esi, [eax + REFRESH_DELAY]			; now holds time when screen should be refreshed next
	
@main_loop:
	invoke PeekMessage, ADDR msg, NULL, 0, 0, PM_REMOVE
	test eax, eax
	.IF !zero?
		cmp msg.message, WM_QUIT
		je @end_main_loop					; quit loop if quit message was encountered
		
		invoke TranslateMessage, ADDR msg
		invoke DispatchMessage, ADDR msg
	.ENDIF

	invoke GetTickCount						; get current time
	cmp esi, eax
	ja @main_loop							; jump if not enough time passed
	lea esi, [eax + REFRESH_DELAY]			; save the next update time
	
	invoke FillRect, edi, ADDR g_world, g_winclass.hbrBackground ; flush background
	invoke UpdateDots, edi					; update dot positions and draw dots
	
	jmp @main_loop
@end_main_loop:
	invoke ReleaseDC, ebx, edi				; release dc..
@force_exit:
	ret
main endp

;------------------------------------------------------------------------------
; Windows callback proc.
;------------------------------------------------------------------------------	
WndProc proc hWnd:DWORD, Msg:DWORD, wParam:DWORD, lParam:DWORD
	cmp Msg, WM_DESTROY						; check if the program wants to die
	je @WM_DESTROY
	
	invoke DefWindowProc, hWnd, Msg, wParam, lParam
	ret
@WM_DESTROY:
	invoke PostQuitMessage, 0
	xor eax, eax
	ret
WndProc endp

CreateDot proc uses edi dot:DWORD
	LOCAL temp:DWORD

	mov edi, dot
	
	; dots lifetime is: MINIMUM_LIFETIME + random(10)
	invoke nrandom, 10
	add eax, MINIMUM_LIFETIME
	mov dword ptr [edi.SINGLE_DOT.lifetime], eax
	
	; creates random x position for dot
	invoke nrandom, g_world.right
	mov temp, eax
	fild dword ptr [temp]
	fstp dword ptr [edi.SINGLE_DOT.position.x]
	
	; creates random y position for dot
	invoke nrandom, g_world.bottom
	mov temp, eax
	fild dword ptr [temp]
	fstp dword ptr [edi.SINGLE_DOT.position.y]

	ret
CreateDot endp

;------------------------------------------------------------------------------
; Creates all dots.
;------------------------------------------------------------------------------	
CreateDots proc uses ebx edi
	mov ebx, AMOUNT_OF_DOTS
	mov edi, OFFSET dots
	
	.REPEAT
		invoke CreateDot, edi
		sub ebx, 1
	.UNTIL zero?
	
	ret
CreateDots endp

;------------------------------------------------------------------------------
; Updates dots position and draws them.
;------------------------------------------------------------------------------	
UpdateDots proc uses ebx esi hDC:HDC
	mov ebx, AMOUNT_OF_DOTS
	mov esi, OFFSET dots
	
	.REPEAT
		cmp dword ptr [esi.SINGLE_DOT.lifetime], 0		; see if the dot is still alive
		je @createdot
		
		; below we check that dot is within the window, otherwise it is created again
		fld dword ptr [esi.SINGLE_DOT.position.x]
		fldz
		fcomip st(0), st(1)
		fwait
		ja @createdot_popfpu
		
		fild dword ptr [g_world.right]
		fcomip st(0), st(1)
		fwait
		jb @createdot_popfpu
		
		fld dword ptr [esi.SINGLE_DOT.position.y]
		fldz
		fcomip st(0), st(1)
		fwait
		ja @createdot_popfpu
		
		fild dword ptr [g_world.bottom]
		fcomip st(0), st(1)
		fwait
		jb @createdot_popfpu
		
		db 0DFh, 0C1h ; ffreep st(1)
		
		sub dword ptr [esi.SINGLE_DOT.lifetime], 1	; decrease dot lifetime with one frame
		
		push 0FFFFFFh	; SetPixel color
		
		fld dword ptr [esi.SINGLE_DOT.position.x]	; get dots x pos
		fsub dword ptr [g_worldcenter.x]			; get vectors x element
		
		fld dword ptr [esi.SINGLE_DOT.position.y]	; get dots y pos
		fsub dword ptr [g_worldcenter.y]			; get vectors y element
		
		push ecx		; SetPixel y position, here we just make some room for the value
		
		fmul dword ptr [g_dotspeed]					; multiply vectors x element with speed
		fadd dword ptr [esi.SINGLE_DOT.position.y]	; update position
		fst dword ptr [esi.SINGLE_DOT.position.y]	; store current position for later use
		fistp dword ptr [esp]						; push the position for SetPixel
		
		push ecx		; SetPixel x position
		
		fmul dword ptr [g_dotspeed]					; multiply vectors y element with speed
		fadd dword ptr [esi.SINGLE_DOT.position.x]	; update position
		fst dword ptr [esi.SINGLE_DOT.position.x]	; store current position for later use
		fistp dword ptr [esp]						; push the position for SetPixel
		
		push hDC		; push dc handle
		call SetPixel	; draw the pixel
		jmp @F
		
	@createdot_popfpu:
		db 0DFh, 0C0h ; ffreep st(0)
	@createdot:
		invoke CreateDot, esi
@@:
		add esi, SIZEOF SINGLE_DOT						; set pointer to next dot
		sub ebx, 1										; decrese counter
	.UNTIL zero?
	
	ret
UpdateDots endp
end start